function [mu, sigma, g, m, a, b, cost] = mcvq_anneal(X, K, J, iterations, varargin)
%[mu, sigma, g, m, a, b, cost] = mcvq(X, K, J, iterations, ...)
%
% Required Arguments:
%    X = data set, data vectors in columns
%    K = number of VQ's
%    J = number of vectors per VQ
%    iterations = number of EM iterations to perform
%
% Optional Arguments:
%    'tie r' = tie the selection of VQ's across all data vectors
%    'tie g' = tie the (approx.) posterior probabilities of VQ's
%    'adapt priors' = use non-uniform priors, that are updated
%       each M-step, for the g's and m's
%    'hyper priors' = use Dirichlet hyper-priors, to encourage
%       the a's and b's to have low and high entropy respectively
%
% Return Values:
%    mu    = 
%    sigma = 
%    g     = approx. posterior prob. of each VQ
%    m     = approx. posterior prob. of each vector of each VQ
%    a     = prior probabilities of VQ's
%    b     = prior probabilities for each vector of each VQ
%    cost  = value of the cost function (free energy) 
%            after each iteration of EM
%
% Author: David Ross
% $Id: mcvq_anneal.m,v 2.5 2004/07/09 15:53:04 dross Exp $

%---------------------------------------------------
% Check the input
%---------------------------------------------------
error(nargchk(4,4,nargin - length(varargin)));

if ndims(X) ~= 2
    error('X matrix must be 2 dimensional')
end 

if K <= 0 | J <= 0
    error('numbers of VQ''s and components must be positive')
end

if iterations <= 0
    error('number of iterations must be positive')
end

% now check the options
g_tied = 0;
r_tied = 0;
adapt_priors = 0;
hyper_priors = 0;

for i = 1:length(varargin)
    if strcmpi(varargin{i}, 'tie g')
        g_tied = 1;
    elseif strcmpi(varargin{i}, 'tie r')
        r_tied = 1;
        g_tied = 1;
    elseif strcmpi(varargin{i}, 'adapt priors')
        adapt_priors = 1;
    elseif strcmpi(varargin{i}, 'hyper priors')
        hyper_priors = 1;
        adapt_priors = 1;
    else
        disp('unknown argument:');
        disp(varargin{i});
        error(['unknown argument']);
    end
end

% if cost is not a desired output, then don't calcualte it
calculate_cost = (nargout >= 7);

% display the options that MCVQ is running with
disp('====================');
disp(sprintf('J=%d, K=%d',J,K));
if r_tied; disp('r tied across cases'); end
if g_tied; disp('g tied across cases'); end
if adapt_priors; disp('adapting priors'); end
if hyper_priors; disp('using hyper-priors'); end


%---------------------------------------------------
% Initialization
%---------------------------------------------------

N = size(X,1); % dimensionality of input vectors
C = size(X,2); % number of training examples

sigma_min = 0.01;  % the minimum possible value of sigma
disp(sprintf('sigma_min = %g',sigma_min));
disp('====================');

%% annealing schedule
%anneal = linspace(C/2,0.5,iterations);       % linearly decreasing
anneal = linspace(3,1,iterations);           % linearly decreasing
%anneal = exp(linspace(log(C),0,iterations)); % exponentially decreasing
%anneal = ones(iterations,1);                 % constant

%% hyper priors
alpha = 0.5 * ones(K,1); % low entropy
beta = 6 * ones(J,1);     % high entropy

%% priors are initialized to samples from the hyper-priors,
% or set to uniform if hyper-priors are not used
if hyper_priors
    a = zeros(N,K);
    for i = 1:N
        a(i,:) = dirichletrnd(alpha)';
    end

    b = zeros(J*K,1);
    for k = 1:K
        b((k-1)*J+1:k*J) = dirichletrnd(beta);
    end
else
    a = ones(N,K) / K;
    b = ones(J*K,1) / J;
end

%% mu's are randomly chosen from the data, using the following algorithm
%
% Selection stores the integers from 1 to C, which are randomly permuted.
% The first J*K elements of selection indicate which training examples
% are chosen as our initial mu's.
%
% Note: if C is less than J*K, then, before permuting, we give selection
% ceil(J*K/C) copies of each integer from 1 to C.
mu = zeros(N, J*K);
selection = repmat(1:C, [1 ceil(J*K/C)]);
selection = selection(randperm(length(selection)));
mu = X(:,selection(1:J*K));
clear selection;

%% sigma initialized randomly in the interval [0.5 1]
sigma = rand(N,J*K) / 2 + 0.5;
% now make sure Sigma is at least sigma_min
sigma(find(sigma < sigma_min)) = sigma_min;

%% m's are updated first, so we can initialize them to 0
m = zeros(J*K,C);

%% g's @TODO these should be sampled from the priors
if g_tied
    g = a;
else
    g = repmat(a,[1 1 C]);
end

%% empty cost vector
cost = [];

%---------------------------------------------------
% Main Loop
%---------------------------------------------------
for it = 1:iterations

    if length(anneal) < it
        anneal(it) = anneal(end);
    end

    %-------------------------------------------------------------------
    % E-Step
    %-------------------------------------------------------------------
    g_old = g;
    g = repmat(log(a), [1 1 C^(~g_tied)]);
    for c = 1:C
        d = log(sigma) + (mu - repmat(X(:,c),[1 J*K])).^2 ./ (2 * sigma.^2);

        %---- Update m ----%
        % big_G is G(N by K) replicated to make it of size (N by JK)
        % If g's are tied, then there is only one big_G for all c's, 
        % otherwise we need to update big_G for each c
        if (~g_tied) | (c == 1)
            big_G = replicate_g(g_old(:,:,c),J);
        end
        
        m(:,c) = log(b) - dot(big_G, d, 1)';
        m_reshaped = reshape(m(:,c), [J K]);
        m_reshaped = m_reshaped / anneal(it);
        m_reshaped = softmax(m_reshaped);
        m(:,c) = reshape(m_reshaped, [J*K 1]);
 
        %---- Update g ----%
        % c_for_g = 1 if g is tied, c if g is untied
        c_for_g = c^(~g_tied);
        
        % @TODO if this is really the E-step, then we shouldn't 
        % recompute d, otherwise if it's the M-step, then this is okay
        g(:,:,c_for_g) = g(:,:,c_for_g) - ...%(1/C)^(~r_tied & g_tied) * ...
            (1/anneal(it)) * ...
            (repmat(m(:,c)',[N 1]) .* d) * kron(eye(K),ones(J,1));
 
    end
    
    % What we want to do is "g = softmax(g')'", but we can't because
    % we have a third index.  The following code does the same thing.
    g = softmax(permute(g,[2 1 3]));
    g = ipermute(g,[2 1 3]);
        

    %-------------------------------------------------------------------
    % M-Step
    %-------------------------------------------------------------------    

    if adapt_priors
        %---- Update a ----%
        % a = mean(g,3); is basically what we're doing, unless
        % hyperpriors are used, which makes things messy

        a = C^(~r_tied & g_tied) * sum(g,3) + ...
            hyper_priors * repmat(alpha'-1, [N 1]);
        a = a / (C^(~r_tied) + hyper_priors * (sum(alpha) - K));
	
        % check to see if any a's are going to zero
        if any(any(a<(1/(10*K))))
            %disp('found a''s that are too small');
            % add a little bit and renormalize
            a = a + (a<(1/(10*K))) * 1/(10*K);
            a = a ./ repmat(sum(a,2), [1 K]);
        end
    
        %---- Update b ----%
        % b = mean(m,2); again this is basically it...
	
        b = sum(m,2) + hyper_priors * repmat(beta-1,[K 1]);
        b = b / (C + hyper_priors * (sum(beta) - J));
	
        % check to see if any b's are going to zero
        if any(b<(1/(10*J)))
            %disp('found b''s that are too small');
            % add a little bit and renormalize
            b = b + (b<(1/(10*J))) * 1/(10*J);
            denom = kron(eye(K), ones(J,J)) * b;
            b = b ./ denom;
        end
    end
    
    %---- Update mu and sigma ----%
    mu_numer = zeros(N, J*K);
    sigma_numer = zeros(N, J*K);
    denom = zeros(N, J*K);

    for c = 1:C
        big_m = repmat(m(:,c)', [N 1]);
        big_X = repmat(X(:,c), [1 J*K]);
        
        % we only need the g factor when g is not tied,
        % otherwise just use a multiplier of 1
        if g_tied
            big_g = 1;
        else
            big_g = replicate_g(g(:,:,c),J);
        end
        
        % accumulate mu
        mu_numer = mu_numer + big_g .* big_m .* big_X;
        
        % accumulate sigma
        sigma_numer = sigma_numer + big_g .* big_m .* (big_X - mu) .^ 2;
        
        % accumulate denominator
        denom = denom + big_g .* big_m;        
    end

    % replace 0's in the denominator with 1's, which is okay since the 
    % corresponding numerator terms should also be 0's
    denom(find(denom ==0)) = 1;
    
    mu = mu_numer ./ denom;
    sigma = sqrt(sigma_numer ./ denom);
    
    % now make sure sigma is at least sigma_min
    sigma(find(sigma < sigma_min)) = sigma_min;

    
    
    %-------------------------------------------------------------------
    % Calculate Free Energy
    %-------------------------------------------------------------------
    if calculate_cost
        
        % note that the C factor only appears when g is tied but r is not
        F = C^(~r_tied & g_tied) * sum(sum(sum(g .* log(g + (g==0))))) ...
            + sum(sum(sum(m .* log(m + (m==0)))));
        
        % d_ijk^c term
        for c = 1:C
            d = log(sigma) + (mu - repmat(X(:,c),[1 J*K])).^2 ./ (2 * sigma.^2);
            
            % replace d with sum_j (m * d)
            d = (repmat(m(:,c)',[N 1]) .* d) * kron(eye(K),ones(J,1));
            
            F = F + sum(sum(g(:,:,c^(~g_tied)) .* d));
        end
        
        % if we're adapting the priors, include these terms    
        % @TODO - we could probably include these even if we aren't
        % adapting the priors
        if adapt_priors
            F = F - (C^(~r_tied & g_tied)) * sum(sum(sum(g,3) .* log(a))) ...
                - sum(sum(m,2) .* log(b));
        end
        
        % if we're using a hyper-prior, include these terms
        if hyper_priors
            F = F -  sum((alpha-1) .* sum(a,1)') ...
                - sum( (beta-1) .* log(sum(reshape(b, [J K]),2)));
        end
        
        cost = [cost F];
        disp(sprintf('iteration %d, cost=%g', it, F));
        
    else
        disp(sprintf('iteration %d',it));
    end
    
    %-------------------------------------------------------------------
    % backup
    %-------------------------------------------------------------------
    save temporary; 
end

%---- Reshape the matricies ----%
% so that j and k *aren't* collapsed into a single matrix dimension
mu = permute(reshape(mu, [N J K]), [2 3 1]);
sigma = permute(reshape(sigma, [N J K]), [2 3 1]);
m = reshape(m, [J K C]);
b = reshape(b, [J K]);

%-----------------------------------------------------------------------
%-----------------------------------------------------------------------
function result = replicate_g(G,J)
% REPLICATE_G this function returns G with columns replicated so now 
% there are J*K of them, rather than just K
K = size(G,2);
result = G(:,kron(1:K,ones(1,J)));


%-----------------------------------------------------------------------
%-----------------------------------------------------------------------
function result = getIndex(j,k,J)
% GETINDEX for a vector of length J*K, this function returns the
% index corresponding to element j,k
result = (k-1)*J + j;
